#include "ScanThread.h"

#include <wx/arrimpl.cpp> // this is a magic incantation which must be done! (that's straight from the docs, by the way)
WX_DEFINE_OBJARRAY(ArrayOfPoints);

#define BW 1

// in the implementation
DEFINE_EVENT_TYPE(IMAGE_UPDATE_EVENT)
DEFINE_EVENT_TYPE(DISPLAY_TEXT_EVENT)
DEFINE_EVENT_TYPE(WAIT_FOR_LASER_EVENT)
DEFINE_EVENT_TYPE(CALIBRATE_EVENT)
DEFINE_EVENT_TYPE(CALIBRATE_MOVE_OBJECT_EVENT)

ScanThread::ScanThread(wxFrame *windowIn, CvCapture *capture) : wxThread(wxTHREAD_JOINABLE)
{
	cvCapture = capture;
	window = windowIn;
	calibrate = false;
	calData = new CalMap();
}

void ScanThread::OnExit()
{
	delete calData;
}

void* ScanThread::Entry()
{
	IplImage *noLaser;	// contains the last frame capture
	IplImage *withLaser, *noLaserCopy, *withLaserCopy, *scanOutput = NULL;

	// load calibration data
	if (calibrate == false)
	{
		if (LoadCalibrationData() == false)
		{
			DisplayText(wxT("\nFailed to read calibration data."));
			return NULL;
		}
	}

	// parse cal file

	while(1)
	{
		if (TestDestroy() == true)
		{
			break;
		}

		// take no-laser image
		SetHorz(LASER_MAX_LEFT/4);
		Sleep(200);
		SetHorz(LASER_MAX_LEFT/2);
		Sleep(200);
		SetHorz(LASER_MAX_LEFT*0.75);
		Sleep(200);
		SetHorz(LASER_MAX_LEFT);


		// take a few images to flush the buffer
		for (int i=0;i<5;i++)
		{
			cvQueryFrame(cvCapture);
		}
		// don't need to release noLaser because it is given by cvQueryFrame
		noLaser = cvQueryFrame(cvCapture);
		noLaserCopy = cvCloneImage(noLaser);

		if (scanOutput != NULL)
		{
		    cvReleaseImage(&scanOutput);
		}

		if (BW)
		{
			CvSize sz = cvSize(noLaser->width & -2, noLaser->height & -2);
			scanOutput = cvCreateImage(sz, 8, 1);
		} else {
			scanOutput = cvCloneImage(noLaser);
		}

		// if we are in calibration mode, wait for the user to move the calibration object

		for (int scanPos=LASER_MAX_LEFT - 1;scanPos>=LASER_MAX_RIGHT;scanPos-=5)
		{
			// begin scan
			SetHorz(scanPos);


			// take a few images to flush the buffer
			for (int i=0;i<1;i++)
			{
				cvQueryFrame(cvCapture);
			}

			withLaser = cvQueryFrame(cvCapture);
			withLaserCopy = cvCloneImage(withLaser);

			if (noLaser != NULL && withLaser != NULL)
			{

				scanOutput = FindLaser(noLaserCopy, withLaserCopy, scanOutput, scanPos);
			}


			cvReleaseImage(&withLaserCopy);

			Yield();




			//cvWaitKey(33);   // (WaitTime);
			//if (m_CamStopped || m_CamPaused) break;
		}
			// and set the new distance
		if (calibrate == true)
		{
			// make sure all events are processed before we change the distance
			Sleep(100);

			WaitForCalibrate();
		}
		cvReleaseImage(&noLaserCopy);;
		DisplayText(wxT(""), -2);

	}
	return NULL;
}

bool  ScanThread::LoadCalibrationData()
{
	// load calibration data
	wxFFile calFile(wxT("calibration.dat"));
	if (calFile.IsOpened() == false)
	{
		return false;
	}
	wxString fullCal;
	calFile.ReadAll(&fullCal);
	calFile.Close();

	// parse data
	wxString distString, angleString, xString, thisLine;
	long distance, angle, xPos;

	/*
	* File Syntax:
	* ------------
	* Distance (cm)>Angle,xPos|Angle,xPos|Angle,xPos
	* Distance (cm)>Angle,xPos|Angle,xPos|Angle,xPos
	* Distance (cm)>Angle,xPos|Angle,xPos|Angle,xPos
	* ------------
	*/

	while (fullCal.length() > 0)
	{
		thisLine = fullCal.BeforeFirst('\n');
		fullCal = fullCal.AfterFirst('\n');

		distString = thisLine.BeforeFirst('>');
		if (distString.ToLong(&distance) == false)
		{
			// conversion error
			return false;
		}

		thisLine = thisLine.AfterFirst('>');

		while (thisLine.length() > 0)
		{
			angleString = thisLine.BeforeFirst(',');
			thisLine = thisLine.AfterFirst(',');
			if (angleString.ToLong(&angle) == false)
			{
				return false;
			}

			xString = thisLine.BeforeFirst('|');
			thisLine = thisLine.AfterFirst('|');
			if (xString.ToLong(&xPos) == false)
			{
				return false;
			}

			// add this point to the data structure
			if (calData->find(angle) == calData->end())
			{
				// this is the first time we have seen a point at this angle
				(*calData)[angle] = new ArrayOfPoints();
			}
			CalMap::iterator iter = calData->find(angle);

			// add this data point to the existing array of points that corresponds to the angle the laser
			// is currently at.
			iter->second->Add(new wxPoint(xPos, distance));
		}
	}
	return true;
}

void ScanThread::DisplayImage(IplImage *frame, wxString str)
{
	wxCommandEvent event(IMAGE_UPDATE_EVENT, GetId());
	// Give it some contents
	event.SetClientData(frame);
	event.SetString(str);
	// Send it
	window->AddPendingEvent(event);
}

void ScanThread::WaitForLaser(bool on)
{

	wxCommandEvent event(WAIT_FOR_LASER_EVENT, GetId());
	if (on == true)
	{
		event.SetString(wxT("\nPlease turn the laser on."));
	} else {
		event.SetString(wxT("\nPlease turn the laser off."));
	}
	wxMutex captureMutex;
	event.SetClientData(&captureMutex);
	window->AddPendingEvent(event);
	// wait for the window to lock
	Sleep(100);
	// hang until the main GUI says OK
	captureMutex.Lock();

	wxCommandEvent event2(DISPLAY_TEXT_EVENT, GetId());

	if (on == true)
	{
		event2.SetString(wxT("\nLaser is on."));
	} else {
		event2.SetString(wxT("\nLaser is off."));
	}

	window->AddPendingEvent(event2);
}

void ScanThread::WaitForCalibrate()
{
	wxCommandEvent event(CALIBRATE_MOVE_OBJECT_EVENT, GetId());
	event.SetString(wxT("\nWaiting for user to move calibration object..."));
	wxMutex captureMutex;
	event.SetClientData(&captureMutex);
	window->AddPendingEvent(event);
	// wait for the window to lock
	Sleep(100);
	// hang until the main GUI says OK
	captureMutex.Lock();

	wxCommandEvent event2(DISPLAY_TEXT_EVENT, GetId());

	event2.SetString(wxT(" done."));

	window->AddPendingEvent(event2);

}

IplImage* ScanThread::FindLaser(IplImage *noLaser, IplImage *withLaser, IplImage *scanOutput, int laserPos)
{

	IplImage *noLaserCopy = cvCloneImage(noLaser);
	IplImage *withLaserCopy = cvCloneImage(withLaser);
	CvSize sz = cvSize(noLaser->width & -2, noLaser->height & -2);
	IplImage *bwNoLaser = cvCreateImage(sz, 8, 1);
	IplImage *bwWithLaser = cvCreateImage(sz, 8, 1);
	IplImage *subImage = cvCreateImage(sz, 8, 1);

	cvSetImageCOI(noLaserCopy, 3);		// select the red channel
	cvSetImageCOI(withLaserCopy, 3);	// select the red channel

	cvCopy(noLaserCopy, bwNoLaser, 0);
	cvCopy(withLaserCopy, bwWithLaser, 0);

	cvSub(bwWithLaser, bwNoLaser, subImage);

	// smooth the image to remove edges
	cvSmooth(subImage, subImage, CV_GAUSSIAN, 9, 9);

	//cvThreshold(subImage, subImage, 55, 255, CV_THRESH_BINARY);
	cvThreshold(subImage, subImage, 40, 255, CV_THRESH_BINARY);
	//cvThreshold(subImage, subImage, 15, 255, CV_THRESH_BINARY);

	// find the center point of the image
	CvMoments *moments = new CvMoments();
	cvMoments(subImage, moments, 1);
	double xPos = cvGetSpatialMoment(moments, 1, 0);
	double yPos = cvGetSpatialMoment(moments, 0, 1);

	int totalNum = cvGetSpatialMoment(moments, 0, 0);

	xPos /= totalNum;
	yPos /= totalNum;

	double x2 = cvGetNormalizedCentralMoment(moments, 2, 0);
	double y2 = cvGetNormalizedCentralMoment(moments, 0, 2);

	//cvCircle(subImage, cvPoint((int)xPos, (int)yPos), 25, cvScalar(255, 255, 255));

	// draw the ellipse (commented out for distance maping)
	//cvEllipse(subImage, cvPoint((int)xPos, (int)yPos), cvSize((int) (x2*300), (int)(y2*300)),0,0,360,cvScalar(255,255,255));

	x2 = (int) (x2*300);
	y2 = (int) (y2*300);
	//DisplayImage(subImage, wxT("Output1"));
	wxString outString = wxT("\nCenter: (");
	outString << xPos;
	outString += wxT(", ");
	outString << yPos;
	outString += wxT(")");

	outString += wxT(" Laser: ");
	outString << laserPos;

	if (calibrate == true && xPos > 0)
	{
		SendCalibrationEvent(int(xPos), laserPos);
	}


	// BAD
	// cubic fit line 3.45872e-5 * x^3 + -0.02485499*x^2 + 5.77422 * x - 111.51846
	// valid from x = 40 cm to 250 cm.  Approx. constant at 330 at x > 250 cm

	// GOOD
	// cubic fit line distance (cm) = 3.67440191e-5 * x^3 - 0.016724404 * x^2 + 2.40843979 * x - 44.8949454
	// where x is pixel value.  valid from approx 169 to 330.


	//double dist = 3.67440191 * 0.000010 * xPos*xPos*xPos - 0.016724404 * xPos*xPos + 2.40843979 * xPos - 44.8949454;
	double dist = PixelToDistance(xPos, laserPos);
	outString += wxT(" -- Distance = ");
	outString << dist;
	outString += wxT("cm");
	DisplayText(outString, dist);

	BwImage subPx(subImage);

	IplImage *outputImage = scanOutput;
	BwImage outPx(outputImage);

	for (int w=0;w<subImage->width;w++)
	{
		for (int h=0;h<subImage->height;h++)
		{
			// for every pixel in myImage, check to see if it is non-zero
			if (subPx[h][w] > 0)
			{
				// this pixel has the laser on it!
				if (BW)
				{
					outPx[h][w] = DistanceToColor(PixelToDistance(w, laserPos));
				} else {

					outPx[h][w*3] = DistanceToColor(PixelToDistance(w, laserPos));
					outPx[h][w*3 + 1] = DistanceToColor(PixelToDistance(w, laserPos));
					outPx[h][w*3 + 2] = DistanceToColor(PixelToDistance(w, laserPos));
				}
			}
		}
	}

	DisplayImage(outputImage, wxT("Output2"));

	cvReleaseImage(&noLaserCopy);
	cvReleaseImage(&withLaserCopy);
	cvReleaseImage(&subImage);
	cvReleaseImage(&bwNoLaser);
	cvReleaseImage(&bwWithLaser);
    delete moments;

	return outputImage;


}

void ScanThread::DisplayText(wxString text, double num)
{
	wxCommandEvent event2(DISPLAY_TEXT_EVENT, GetId());
	event2.SetString(text);
	double *num2 = new double(num);
	event2.SetClientData(num2);
	window->AddPendingEvent(event2);
}

double ScanThread::PixelToDistance(int xPos, int laserPos)
{
	//return 3.67440191 * 0.000010 * xPos*xPos*xPos - 0.016724404 * xPos*xPos + 2.40843979 * xPos - 44.8949454;

	// convert pixel location and laser position into a distance using calibration data

	// first, identify the laser position in the hash map

    if (calData->find(laserPos) == calData->end())
    {
        // no data
        return -1;
    }
	ArrayOfPoints *pointArray = (*calData)[laserPos];


	// search the array of points for the two that bracket this x position

	int lower = -1, higher = 10000, lowerDist, higherDist;

	for (int i=0;i<int(pointArray->GetCount());i++)
	{
		// wxPoint(xPos, distance)
		if (pointArray->Item(i)->x < xPos && pointArray->Item(i)->x > lower)
		{
			lower = pointArray->Item(i)->x;
			lowerDist = pointArray->Item(i)->y;
		}

		if (pointArray->Item(i)->x > xPos && pointArray->Item(i)->x < higher)
		{
			higher = pointArray->Item(i)->x;
			higherDist = pointArray->Item(i)->y;
		}
	}

	// check to make sure we found data
	if (lower == -1 || higher == 10000)
	{
		// no calibration data!
		return -1;
	}

	// we have calibration data

	// run a linear interpolation on the two points
	float temp = float(higherDist-lowerDist)/float(higher-lower);
	float temp2 = (xPos - lower) * ( temp ) + lowerDist;

	return temp2;


}

int ScanThread::DistanceToColor(double dist)
{
	// max distance ~= 330
	// min distance ~= 30
	int max = 110;
	int min = 75;
	if (dist > max)
	{
		return 0;
	}
	if (dist < min)
	{
		return 255;
	}
	int out = 255 - 255 * float(dist - min) / float(max - min);
	return out;
}

void ScanThread::SendCalibrationEvent(int xPos, int laserPos)
{
	wxCommandEvent event(CALIBRATE_EVENT, GetId());
	event.SetInt(xPos);
	event.SetExtraLong(laserPos);

	window->AddPendingEvent(event);
}
